"
 (C) 2012 by Holger Hans Peter Freyther
 All Rights Reserved

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.

 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"

LogicalChannel extend [
    sendGSM: aMsg [
        <category: '*-OpenBSC-Test'>
        self sendGSM: aMsg sapi: 0.
    ]

    sendGSM: aMsg sapi: aSapi [
        <category: '*-OpenBSC-Test'>

        (self sapiIsEstabliashed: aSapi)
            ifTrue:  [self sendData: aMsg on: aSapi]
            ifFalse: [self establish: aMsg on: aSapi].
    ]
]

Object subclass: LogicalChannelWrapper [
    | sapi0 sapi3 lchan |
    <comment: 'I am wrapping a LogicalChannel and provide SAPI
    access to it.'>

    LogicalChannelWrapper class >> initWith: aLchan [
        <category: 'creation'>
        ^ self new
            lchan: aLchan; yourself
    ]

    lchan: aLchan [
        <category: 'creation'>
        lchan := aLchan.
        sapi0 := SharedQueue new.
        sapi3 := SharedQueue new.
        lchan onDataRequest: [:msg :sapi | 
            sapi = 0
                ifTrue: [sapi0 nextPut: msg].
            sapi = 3
                ifTrue: [sapi3 nextPut: msg].
        ].      
    ]

    nextSapi0Msg [
        <category: 'msg'>
        ^ sapi0 next
    ]

    nextSapi3Msg [
        <category: 'msg'>
    ]

    sendGSM: aGSM [
        <category: 'sending'>
        lchan sendGSM: aGSM
    ]

    releaseAllSapis [
        <category: 'release'>

        Transcript nextPutAll: 'Releasing all SAPIs of the channel'; nl.

        (lchan sapiIsEstabliashed: 0)
            ifTrue: [lchan releaseSapi: 0].
        (lchan sapiIsEstabliashed: 3)
            ifTrue: [lchan releaseSapi: 3].
    ]

    onReleaseReqCB: aCb [
        <category: 'release'>
        lchan onReleaseReqCB: aCb
    ]

    cancel [
        <category: 'release'>

        sapi0 nextPut: nil.
        sapi3 nextPut: nil.
    ]
]

Object subclass: OpenBSCTest [
    | bts testFailed |
    <category: 'OpenBSC-Test'>
    <comment: 'I help in dealing with setup and teardown of a test'>

    OpenBSCTest class >> initWith: aBTS [
        <category: 'creation'>
        ^ self new
            bts: aBTS; yourself.
    ]

    bts: aBTS [
        <category: 'creation'>
        bts := aBTS
    ]

    createAndConnectBTS [
        <category: 'bts'>
        bts := BTS new.
        bts connect: 'localhost'.
        bts waitForBTSReady.
    ]

    stopBts [
        <category: 'bts'>
        bts stop.
    ]

    requireChannel: aType random: aMask [
        | ra rsl lchan |
        <category: 'bts'>

        "The RA we will wait for.."
        ra := aType bitOr: (Random between: 0 and: aMask).

        rsl := RSLChannelRequired new.
        rsl channelNumber: RSLChannelNumber ccchRach.
        rsl requestReference: {ra. 42. 20} asRSLAttributeData.
        rsl accessDelay: #(23) asRSLAttributeData.

        lchan := bts waitForChannel: rsl toMessage with: ra.
        ^ LogicalChannelWrapper initWith: lchan.
    ]

    requireAnyChannel [
        <category: 'bts'>
        " Only use four bit for random to work with both necis"
        ^ self requireChannel: 2r0 random: 2r1111
    ]

    requireEmergencyChannel [
        <category: 'bts'>
        ^ self requireChannel: 2r10100000 random: 2r11111
    ]

    assert: aBoolean message: aMessage [
        <category: 'verifying'>
        aBoolean ifTrue: [^self].

        Transcript nextPutAll: 'TEST: Failure with ', aMessage; nl.
        testFailed := true.
    ]

    deny: aBoolean message: aMessage [
        <category: 'verifying'>
        self assert: aBoolean not message: aMessage.
    ]

    failed [
        <category: 'verifying'>
        ^ testFailed
    ]
]

